/**
 *
 * Project: Eight Track, File: GuildMusicHandler.java
 *
 * Copyright 2019-2020 Locusworks LLC.
 * All rights reserved. Federal copyright law prohibits unauthorized reproduction by
 * any means and imposes fines up to $25,000 for violation.  No part of this material
 * may be reproduced, transmitted, transcribed, stored in a retrieval system, copied,
 * modified, duplicated, adapted or translated into another program language in any
 * form or by any means, electronic, mechanical, photocopying, recording, or
 * otherwise, without the prior written permission from Locusworks. Locusworks
 * affirms that Eight-Track(R) software and data is subject to United States
 * Government Purpose Rights.  Contact Locusworks, 1313 Lawnview Drive
 * Forney TX 75126, (802) 488-0438, for commercial licensing opportunities.
 *
 * IN NO EVENT SHALL LOCUSWORKS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF LOCUSWORKS HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  NO RESPONSIBILITY IS ASSUMED BY
 * LOCUSWORKS FOR ITS USE, OR FOR ANY INFRINGEMENTS OF PATENTS OR OTHER RIGHTS OF
 * THIRD PARTIES RESULTING FROM ITS USE.  LOCUSWORKS SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE AND
 * ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
 * IS". LOCUSWORKS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 */
package net.locusworks.discord.eighttrack.handlers;

import java.time.OffsetDateTime;
import java.util.List;
import java.util.Random;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import com.sedmelluq.discord.lavaplayer.track.AudioTrack;

import net.dv8tion.jda.api.EmbedBuilder;
import net.dv8tion.jda.api.Permission;
import net.dv8tion.jda.api.entities.ChannelType;
import net.dv8tion.jda.api.entities.GuildChannel;
import net.dv8tion.jda.api.entities.MessageEmbed;
import net.dv8tion.jda.api.entities.VoiceChannel;
import net.dv8tion.jda.api.events.message.guild.GuildMessageReceivedEvent;
import net.dv8tion.jda.api.managers.AudioManager;
import net.locusworks.discord.eighttrack.audio.EightTrackAudioSendHandler;
import net.locusworks.discord.eighttrack.audio.TrackManager;
import net.locusworks.discord.eighttrack.database.entities.GuildPlaylist;
import net.locusworks.discord.eighttrack.database.entities.GuildSong;
import net.locusworks.discord.eighttrack.services.RepositoryService;
import net.locusworks.logger.ApplicationLogger;
import net.locusworks.logger.ApplicationLoggerFactory;

public class GuildMusicHandler {

  private ApplicationLogger logger;
  private AtomicBoolean playing;
  private Long voiceChannelId;

  private OffsetDateTime lastPlayed;

  private Consumer<Long> callback;
  private RepositoryService guildSongRepoService;
  private long guildId;

  private GuildPlaylist currentPlaylist;

  private TrackManager trackManager;

  public GuildMusicHandler(long guildId, RepositoryService guildSongRepoService) {
    this.logger = ApplicationLoggerFactory.getLogger(GuildMusicHandler.class);
    this.playing = new AtomicBoolean(false);
    this.lastPlayed = OffsetDateTime.now();
    this.guildSongRepoService = guildSongRepoService;
    this.guildId = guildId;
    this.trackManager = new TrackManager();
  }

  public Long getCurrentVoiceChannelId() {
    return voiceChannelId;
  }

  public OffsetDateTime getLastPlayed() {
    return lastPlayed;
  }

  public void accept(Consumer<Long> callback) {
    this.callback = callback;
  }

  public void isPlaying(boolean playing) {
    this.playing.set(playing);
  }

  public void upNext(GuildMessageReceivedEvent event) throws Exception {
    upNext(event, null);
  }

  public void upNext(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (commands != null && !commands.isEmpty()) {
      String uuid = commands.remove(0);
      boolean foundSong = findSong(uuid);
      if (!foundSong) {
        event.getChannel().sendMessageFormat("Unable to find song with identifier: %s", uuid).queue();
        return;
      }
    }
    if (!trackManager.hasTracks()) {
      loadRandomSong();
    }
    AudioTrack track = trackManager.play();
    MessageEmbed embed = new EmbedBuilder()
        .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
        .setTitle("Next Up:")
        .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
        .setTimestamp(OffsetDateTime.now())
        .build();
    event.getChannel().sendMessage(embed).queue();
  }

  public void repeat(GuildMessageReceivedEvent event) throws Exception {
    next(event);
    trackManager.setFinishedCallback((ater) -> {
      if (!playing.get()) return;
      try {
        next(event);
      } catch (Exception e) {
        logger.error(e);
      }
    });
  }

  public void next(GuildMessageReceivedEvent event) throws Exception {
    next(event, null);
  }

  public void next(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (!playing.get()) {
      play(event, commands);
      playing.set(true);
      return;
    }
    
    GuildChannel gc = event.getGuild().getGuildChannelById(ChannelType.VOICE, voiceChannelId);
    if (gc != null && gc.getMembers().size() == 1) {
      event.getChannel().sendMessage("Going silent since no one is currently listening to the channel").queue();
      event.getGuild().getAudioManager().closeAudioConnection();
      if (callback != null) callback.accept(event.getGuild().getIdLong());
      return;
    }

    playing.set(false);
    stop(event, true, true);
    play(event, commands);
    playing.set(true);
  }

  public void stop(GuildMessageReceivedEvent event) {
    stop(event, false, false);
  }

  public void stop(GuildMessageReceivedEvent event, boolean stoppedFromRepeat, boolean stoppedFromNext) {
    trackManager.stopTrack();
    if (!stoppedFromNext) {
      voiceChannelId = null;
    }
    if (!stoppedFromRepeat) {
      currentPlaylist = null;
    }
  }

  public void play(GuildMessageReceivedEvent event) throws Exception {
    play(event, null);
  }

  public void play(GuildMessageReceivedEvent event, List<String> commands) throws Exception {
    if (playing.get()) return;

    VoiceChannel vc = event.getMember().getVoiceState().getChannel();
    if (vc == null) {
      event.getChannel().sendMessage(String.format("<@%s> you are not in a voice channel to play music", event.getMember().getId())).queue();
      return;
    }

    if (!event.getGuild().getSelfMember().hasPermission(vc, Permission.VOICE_CONNECT, Permission.VOICE_SPEAK)) {
      event.getChannel().sendMessage(String.format("<@%s>, I cannot play my music in channel %s as I dont have permission to", 
          event.getMember().getId(), vc.getName())).queue();
      return;
    }

    Long tmpId = vc.getIdLong();
    if (voiceChannelId != null && voiceChannelId != tmpId) {
      String channelName = event.getGuild().getVoiceChannelById(voiceChannelId).getName();
      event.getChannel().sendMessage(String.format("<@%s> I am currently playing in channel %s. Cannot switch channels until I am stopped.", 
          event.getMember().getId(), channelName)).queue();
      return;
    }

    voiceChannelId = tmpId;

    AudioManager manager = event.getGuild().getAudioManager();
    manager.openAudioConnection(vc);

    manager.setSendingHandler(new EightTrackAudioSendHandler(trackManager));

    if (commands != null && !commands.isEmpty()) {
      String uuid = commands.remove(0);
      boolean foundSong = findSong(uuid);
      if (!foundSong) {
        event.getChannel().sendMessageFormat("Unable to find song with identifier: %s", uuid).queue();
        return;
      }
      stop(event, true, true);
    }

    if (!trackManager.hasTracks()) {
      loadRandomSong();
    }

    if (trackManager.hasTracks()) {
      lastPlayed = OffsetDateTime.now(); 

      AudioTrack track = trackManager.play();
      if (track == null) {
        logger.warn("Unable to find track: " + trackManager.getCurrentItem()) ;
        return;
      }

      MessageEmbed embed = new EmbedBuilder()
          .setAuthor(event.getMember().getEffectiveName(), null, event.getAuthor().getAvatarUrl())
          .setTitle("Now Playing:")
          .setDescription(String.format("**%s** by __%s__", track.getInfo().title, track.getInfo().author))
          .setTimestamp(OffsetDateTime.now())
          .build();

      event.getChannel().sendMessage(embed).queue();
      trackManager.playTrack(track);
    }
  }

  private void loadRandomSong() throws Exception {
    List<GuildSong> gsList = null;

    if (currentPlaylist == null) {
      gsList = guildSongRepoService.getGuildSongRepo().findByGuild(guildId);
    } else {
      gsList = guildSongRepoService.getGuildPlaylistSongRepo().findByGuildPlaylist(currentPlaylist).stream().map(gpl -> gpl.getGuildSong()).collect(Collectors.toList());
    }
    if (gsList == null || gsList.isEmpty()) return;

    Random random = new Random(System.currentTimeMillis());
    int item = random.nextInt(gsList.size());

    GuildSong song = gsList.get(item);
    trackManager.queueNext(song.getSong().getFilePath());
  }

  private boolean findSong(String uuid) throws Exception {
    GuildSong gs = guildSongRepoService.getGuildSongRepo().findByUuid(uuid);
    if (gs == null) return false;

    trackManager.queueNext(gs.getSong().getFilePath());
    return true;
  }

  public void setCurrentPlaylist(GuildPlaylist currentPlaylist) {
    this.currentPlaylist = currentPlaylist;
  }
  
  public TrackManager getTrackManager() {
    return this.trackManager;
  }

}
